@@ -19,6 +19,9 @@ static void      ddm_data_model_class_init          (DDMDataModelClass  *klass);
 static void      ddm_data_model_dispose             (GObject              *object);
 static void      ddm_data_model_finalize            (GObject              *object);
 
+static void ddm_data_model_flush_internal(DDMDataModel *model,
+                                          gboolean      shutting_down);
+
 struct _DDMDataModel {
     GObject parent;
 
@@ -570,6 +573,9 @@ model_reset_changed_resource_foreach (gpointer key,
 void
 ddm_data_model_reset (DDMDataModel *model)
 {
+    /* Pending work items might expect the model to be in the pre-reset state */
+    ddm_data_model_flush_internal(model, TRUE);
+    
     g_hash_table_foreach_remove(model->resources, model_reset_foreach, NULL);
     g_hash_table_foreach_remove(model->changed_resources, model_reset_changed_resource_foreach, NULL);
 
@@ -655,11 +661,13 @@ compare_work_items(gconstpointer  a,
     return (serial_a < serial_b) ? -1 : (serial_a == serial_b ? 0 : 1);
 }
 
-void
-_ddm_data_model_add_work_item (DDMDataModel    *model,
-                               DDMWorkItem     *item)
+static void
+_ddm_data_model_add_work_item_internal (DDMDataModel    *model,
+                                        DDMWorkItem     *item,
+                                        gboolean         schedule_flush)
 {
     GList *l;
+    gint64 serial = _ddm_work_item_get_min_serial(item);
     
     _ddm_work_item_ref(item);
 
@@ -676,7 +684,7 @@ _ddm_data_model_add_work_item (DDMDataModel    *model,
      * because we insert after on the tie-break to avoid reordering
      * when we pull things off the front.
      */
-    if (_ddm_work_item_get_min_serial(item) == -1) {
+    if (serial == -1) {
         for (l = model->work_items->head; l; l = l->next) {
             if (compare_work_items(l->data, item, NULL) > 0)
                 break;
@@ -698,10 +706,22 @@ _ddm_data_model_add_work_item (DDMDataModel    *model,
         else
             g_queue_insert_after(model->work_items, l, item);
     }
-    
 
-    g_debug("Scheduling flush due to work item");
-    ddm_data_model_schedule_flush(model);
+    if (schedule_flush && model->flush_idle == 0 && serial <= model->max_answered_query_serial) {
+        g_debug("Scheduling flush due to work item %s with min_serial "
+                "%" G_GINT64_MODIFIER "d"
+                "; max_answered=%" G_GINT64_MODIFIER "d",
+                _ddm_work_item_get_id_string(item),
+                serial, model->max_answered_query_serial);
+        ddm_data_model_schedule_flush(model);
+    }
+}
+
+void
+_ddm_data_model_add_work_item (DDMDataModel    *model,
+                               DDMWorkItem     *item)
+{
+    _ddm_data_model_add_work_item_internal(model, item, TRUE);
 }
 
 gboolean
@@ -780,7 +800,8 @@ data_model_flush_notifications(DDMDataModel *model)
 }
 
 static void
-data_model_flush_work_items(DDMDataModel *model)
+data_model_flush_work_items(DDMDataModel *model,
+                            gboolean      shutting_down)
 {
     GList *items;
     GList *l;
@@ -820,8 +841,16 @@ data_model_flush_work_items(DDMDataModel *model)
     for (l = items; l; l = l->next) {
         DDMWorkItem *item = l->data;
 
-        if (!_ddm_work_item_process(item))
-            _ddm_data_model_add_work_item(model, item);
+        if (shutting_down)
+            _ddm_work_item_cancel(item);
+        else if (!_ddm_work_item_process(item)) {
+            /* Re-adding the the same item back to the queue should not schedule
+             * a flush, because being re-added means a) what you are waiting for
+             * has not happened or b) you've requested something new from the
+             * server and you are now waiting for that.
+             */
+            _ddm_data_model_add_work_item_internal(model, item, FALSE);
+        }
 
         _ddm_work_item_unref(item);
     }
@@ -829,23 +858,33 @@ data_model_flush_work_items(DDMDataModel *model)
     g_list_free(items);
 }
 
+static void
+ddm_data_model_flush_internal(DDMDataModel *model,
+                              gboolean      shutting_down)
+{
+    if (shutting_down)
+        g_debug("Flushing Data Model on shut down");
+    else
+        g_debug("Flushing Data Model");
+
+    if (model->backend->flush)
+        model->backend->flush(model, model->backend_data);
+
+    data_model_flush_rules(model);
+    data_model_flush_notifications(model);
+    data_model_flush_work_items(model, shutting_down);
+}
+
 void
 ddm_data_model_flush(DDMDataModel *model)
 {
     if (model->flush_idle == 0)
         return;
 
-    g_debug("Flushing Data Model");
-
     g_source_remove(model->flush_idle);
     model->flush_idle = 0;
 
-    if (model->backend->flush)
-        model->backend->flush(model, model->backend_data);
-
-    data_model_flush_rules(model);
-    data_model_flush_notifications(model);
-    data_model_flush_work_items(model);
+    ddm_data_model_flush_internal(model, FALSE);
 }
 
 void
@@ -891,7 +930,6 @@ ddm_data_model_add_rule (DDMDataModel       *model,
     }
     
     source_rules = g_hash_table_lookup(model->rules_by_source, source_class_id);
-    source_rules = g_slist_prepend(source_rules, rule);
     if (source_rules) {
         source_rules = g_slist_prepend(source_rules, rule);
         g_hash_table_insert(model->rules_by_source,
@@ -1001,6 +1039,21 @@ _ddm_data_model_query_answered (DDMDataModel *model,
                                 DDMDataQuery *query)
 {
     gint64 serial = _ddm_data_query_get_serial(query);
+
+    /* There may be waiting work items that are satisfied because this query is now answered. 
+     */
+    if (model->flush_idle == 0 && model->work_items->length > 0)  {
+        gint64 waiting_serial = _ddm_work_item_get_min_serial(model->work_items->head->data);
+        if (waiting_serial <= serial) {
+            g_debug("Scheduling flush because waiting work item %s with min_serial %" G_GINT64_MODIFIER "d "
+                    "potentially satisfied by %s",
+                    _ddm_work_item_get_id_string(model->work_items->head->data),
+                    waiting_serial,
+                    ddm_data_query_get_id_string(query));
+            ddm_data_model_schedule_flush(model);
+        }
+    }
+    
     if (serial > model->max_answered_query_serial)
         model->max_answered_query_serial = serial;
 }
